// =====================================
// 1. Imports and Configurations
// =====================================
import 'dotenv/config';
import express from 'express';
import expressWs from 'express-ws';
import { twiml } from 'twilio';
import { ElevenLabsClient } from 'elevenlabs';
import WebSocket from 'ws';
import { spawn } from 'child_process';
import { Readable } from 'stream';
import path from 'path';
import fs from 'fs/promises';
import OpenAI from 'openai';
import { createClient, LiveTranscriptionEvents, DeepgramClient, DeepgramError } from '@deepgram/sdk';
import Groq from 'groq-sdk';
import { Buffer } from 'buffer';
import { EventEmitter } from 'events';
import webhookRouter from './webhook';

// Global variable for called number
let calledNumber = '';

// Configuration
type OpenAIVoice = "alloy" | "echo" | "fable" | "onyx" | "nova" | "shimmer";

interface AppConfig {
  enable_interruptions: boolean;
  wordsToInterrupt: string[];
  wordToInterruptEnable: boolean;
  numberofwordstointeruptenable: boolean;
  words_to_interrupt: number;
  silenceTimeoutSeconds: number;
  response_delay: number;
  llm_request_delay: number;
  clear_message_delay: number;
  transcriber_provider: string;
  transcriber_model: string;
  model_provider: string;
  model: string;
  temperature: number;
  top_p: number;
  max_tokens: number;
  voice_provider: string;
  voice_id: string | OpenAIVoice;
  stability: number;
  clarity_boost: number;
  speed: number;
  maximum_call_duration: number;
  maximum_call_duration_enable: boolean;
  hipaa_compliance: boolean;
  recording_enabled: boolean;
  server_url: string;
  voicemail_message: string;
  end_call_message: string;
}

const appConfig: AppConfig = {
  enable_interruptions: false, // Default value, will be overridden by assistant settings if available
  wordsToInterrupt: ['stop', 'wait', 'hold on'],
  wordToInterruptEnable: false,
  numberofwordstointeruptenable: true,
  words_to_interrupt: 3, // Default value, will be overridden by assistant settings if available
  silenceTimeoutSeconds: 30, // Default value, will be overridden by assistant settings if available
  response_delay: 0.1, // Default value, will be overridden by assistant settings if available
  llm_request_delay: 0.1, // Default value, will be overridden by assistant settings if available
  clear_message_delay: 500, // Default value, will be overridden by assistant settings if available
  transcriber_provider: 'deepgram', // Default value, will be overridden by assistant settings if available
  transcriber_model: 'nova-2', // Default value, will be overridden by assistant settings if available
  model_provider: 'openai', // Default value, will be overridden by assistant settings if available
  model: 'gpt-3.5-turbo', // Default value, will be overridden by assistant settings if available
  temperature: 0.7, // Default value, will be overridden by assistant settings if available
  top_p: 1, // Default value, will be overridden by assistant settings if available
  max_tokens: 150, // Default value, will be overridden by assistant settings if available
  voice_provider: 'elevenlabs', // Default value, will be overridden by assistant settings if available
  voice_id: 'BJStPCleYBG69brKsxqv', // Default value, will be overridden by assistant settings if available
  stability: 0.5, // Default value for ElevenLabs, will be overridden by assistant settings if available
  clarity_boost: 0.75, // Default value for ElevenLabs, will be overridden by assistant settings if available
  speed: 1.0, // Default value, will be overridden by assistant settings if available
  maximum_call_duration: 600, // Default value, 600 seconds (10 minutes)
  maximum_call_duration_enable: false, // Default value, will be overridden by assistant settings if available
  hipaa_compliance: true, // to implement
  recording_enabled: true, // to implement
  server_url: "http://", // Default value, should be updated with actual server URL. to implement
  voicemail_message: "You have reached our voicemail. Please leave a message after the tone.", // to implement
  end_call_message: "Goodbye" // to implement
};

async function getAssistantSettings(phoneNumber: string) {
  log(`Fetching assistant settings for phone number: ${phoneNumber}`);
  const cachePath = path.join(__dirname, '..', 'cache_settings', 'assistantSettings.json');
  try {
    const fileContent = await fs.readFile(cachePath, 'utf8');
    const assistants = JSON.parse(fileContent);
    const settings = assistants.find((a: any) => a.phone_number === phoneNumber);
    if (settings) {
      log(`Found assistant settings for ${phoneNumber}:`, settings);
      
      // Update appConfig with the fetched settings
      if (settings.transcriber_provider) {
        appConfig.transcriber_provider = settings.transcriber_provider;
      }
      if (settings.transcriber_model) {
        appConfig.transcriber_model = settings.transcriber_model;
      }
      if (settings.silence_timeout) {
        appConfig.silenceTimeoutSeconds = parseInt(settings.silence_timeout);
      }
      if (settings.model_provider) {
        appConfig.model_provider = settings.model_provider;
      }
      if (settings.model) {
        appConfig.model = settings.model;
      }
      if (settings.temperature) {
        appConfig.temperature = parseFloat(settings.temperature);
      }
      if (settings.top_p) {
        appConfig.top_p = parseFloat(settings.top_p);
      }
      if (settings.max_tokens) {
        appConfig.max_tokens = parseInt(settings.max_tokens);
      }
      if (settings.voice_provider) {
        appConfig.voice_provider = settings.voice_provider;
      }
      if (settings.voice_id) {
        appConfig.voice_id = settings.voice_id;
      }
      if (settings.stability && settings.voice_provider === 'elevenlabs') {
        appConfig.stability = parseFloat(settings.stability);
      }
      if (settings.clarity_boost && settings.voice_provider === 'elevenlabs') {
        appConfig.clarity_boost = parseFloat(settings.clarity_boost);
      }
      if (settings.speed) {
        appConfig.speed = parseFloat(settings.speed);
      }
      if (settings.response_delay) {
         appConfig.response_delay = parseFloat(settings.response_delay);
      }
      if (settings.llm_request_delay) {
         appConfig.llm_request_delay = parseFloat(settings.llm_request_delay);
      }
      if (settings.clear_message_delay) {
         appConfig.clear_message_delay = parseFloat(settings.clear_message_delay);
      }
      if (settings.enable_interruptions) {
       appConfig.enable_interruptions = Boolean(settings.enable_interruptions);
      }
      if (settings.words_to_interrupt) {
         appConfig.words_to_interrupt = parseInt(settings.words_to_interrupt);
      }
      
      return settings;
    } else {
      log(`No assistant settings found for phone number: ${phoneNumber}`);
      return null;
    }
  } catch (error) {
    log('Error reading assistantSettings.json:', error);
    return null;
  }
}

// The fetchAssistantSettings function has been moved to appConfig.ts
// Import and use it from there when needed
import { fetchAssistantSettings } from './config/appConfig';

// Function to set the LLM API type based on the model provider
function setLLMApiType(modelProvider: string) {
  switch (modelProvider.toLowerCase()) {
    case 'openai':
      CURRENT_LLM_API = LLMApiType.OpenAIChat;
      break;
    case 'groq':
      CURRENT_LLM_API = LLMApiType.Groq;
      break;
    default:
      log('Unknown model provider, defaulting to OpenAIChat', { provider: modelProvider });
      CURRENT_LLM_API = LLMApiType.OpenAIChat;
  }
  log('Set LLM API Type', { provider: modelProvider, apiType: CURRENT_LLM_API });
}


// Initialize API clients and configurations
const { app } = expressWs(express());
const PORT: number = parseInt(process.env.PORT || '3000', 10);
const elevenlabs = new ElevenLabsClient({ apiKey: process.env.ELEVENLABS_API_KEY });
const openai = new OpenAI({ apiKey: process.env.OPENAI_API_KEY });
const deepgramApiKey = process.env.DEEPGRAM_API_KEY;
if (!deepgramApiKey) {
  throw new Error('DEEPGRAM_API_KEY is not set in the environment variables');
}
const deepgram = createClient(deepgramApiKey);
const groq = new Groq({apiKey: process.env.GROQ_API_KEY});

let isFirstMessage = true;
let silenceTimer: NodeJS.Timeout | null = null;
let ws: WebSocket;
let isAssistantSpeaking = false;
let isWaitingForAudioComplete = false;

const RIME_AI_CONFIG = {
  speaker: process.env.RIME_AI_SPEAKER || 'default_speaker_name',
  modelId: process.env.RIME_AI_MODEL || 'mist',
  reduceLatency: process.env.RIME_AI_REDUCE_LATENCY === 'true'
};

// =====================================
// 2. Type Definitions and Interfaces
// =====================================
enum LLMApiType {
  OpenAIAssistant,
  OpenAIChat,
  Groq
}

enum TTSProvider {
  ElevenLabs,
  OpenAI,
  Deepgram,
  RimeAI
}

enum CommunicationMethod {
  NonStreaming,
  Streaming,
  WebSocket
}

type AssistantStreamEvent = {
  event: string;
  data: any;
};

interface TwilioMessage {
  event: string;
  streamSid?: string;
  sequenceNumber?: string;
  start?: {
    streamSid: string;
    callSid: string;
    accountSid: string;
    tracks: string[];
    customParameters: Record<string, string>;
    mediaFormat: {
      encoding: string;
      sampleRate: number;
      channels: number;
    };
  };
  media?: {
    track: string;
    chunk: string;
    timestamp: string;
    payload: string;
  };
  stop?: {
    accountSid: string;
    callSid: string;
  };
  dtmf?: {
    track: string;
    digit: string;
  };
  mark?: {
    name: string;
  };
}

interface ConversationMessage {
  role: 'user' | 'assistant';
  content: string;
}

type ChatMessage = {
  role: 'system' | 'user' | 'assistant';
  content: string;
};

type MessageContentText = {
  type: 'text';
  text: {
    value: string;
    annotations: any[];
  };
};

type MessageContentImage = {
  type: 'image_file';
  image_file: {
    file_id: string;
  };
};

type MessageContent = MessageContentText | MessageContentImage;

// =====================================
// 3. Global Variables and Constants
// =====================================
let isSpeaking = false;
const interruptionEmitter = new EventEmitter();
let mediaSequenceNumber = 0;
let CURRENT_LLM_API: LLMApiType = LLMApiType.Groq;
let isProcessing = false;
const CURRENT_TTS_PROVIDER: TTSProvider = TTSProvider.ElevenLabs;
const COMMUNICATION_METHOD: CommunicationMethod = CommunicationMethod.Streaming;
const ffmpegPath = path.join(__dirname, '..', 'third-party', 'ffmpeg', 'ffmpeg');

const voiceId = 'BJStPCleYBG69brKsxqv';
const modelId = 'eleven_turbo_v2_5';
let currentThreadId: string | null = null;
const USE_STREAMING = false;

const SYSTEM_MESSAGE = `You are a conversational assistant. Use short, conversational responses as if you're having a live conversation. Your response should be under 20 words. Do not respond with any code, only conversation.\nBlinks Media: Digital Services for Local Businesses\nAbout Us: Blinks Media, based in Dallas, Texas (business phone number: 972-627-4487), is your go-to digital services partner. We specialize in web design, online reputation management, and targeted online marketing strategies. Our mission is to enhance your business's online presence, build a strong digital reputation, and drive growth beyond geographical boundaries. We also offer a free website mockup to showcase our design skills!\nYour Role: Mia, Appointment Booking Assistant\nCommunication Style: Be conversational, friendly, and empathetic. Avoid sounding robotic or generic. Make the prospect feel at ease before diving into questions. Gently guide the conversation back to booking appointments if it strays. Keep our process under wraps.\nResponse Guidelines:\nBe concise: Respond succinctly, addressing one topic at a time. Keep responses under 160 characters. Avoid repetitive answers and unnecessary details.\nEmbrace variety: Use diverse language and rephrasing to enhance clarity without repeating content. Use everyday language, making the chat feel like talking to a friend.\nBe proactive: Lead the conversation, often wrapping up with a question or next-step suggestion. Avoid multiple questions in a single response.\nGet clarity: If the user only partially answers a question or if the answer is unclear, keep asking to get clarity.\nAdapt and guess: Try to understand transcripts that may contain transcription errors. Avoid mentioning \"transcription error\" in the response.\nStay on topic: If the conversation veers off course, politely guide it back. Don’t restart; continue where you left off.\nStay in character: Keep conversations within your role's scope. Guide them back creatively without repeating.\nEnsure fluid dialogue: Respond in a role-appropriate, direct manner to maintain a smooth conversation flow. Use a colloquial way of referring to the date (like 'next Friday', 'tomorrow').\nUse approved content only: Never fabricate responses. If unsure, admit it and steer the conversation back to relevant topics.\nAvoid emojis and promotions: Do not use emojis. Only discuss services if the recipient has shown interest. Confirm interest before offering additional help.\nPersonal touch: Ask for the user's first name for a personal touch but use it sparingly. Avoid overusing names to maintain a natural tone.\nNon-service questions: For topics unrelated to Blinks Media’s services, acknowledge the question and suggest online resources before refocusing on our services. For example: \"Interesting question! While it's outside our focus, online search might help. How can we assist with our digital services?\"\nSpam handling: Politely acknowledge spam but concisely discourage further engagement. Example: \"Thanks for reaching out. We're not interested in new opportunities right now. Best wishes!\"\nOne question at a time: Ask only one question at a time. Do not pack multiple topics into one response.\nStep-by-Step Guide:\nStep 1: Greeting\nIntroduce yourself as Mia from Blinks Media.\nKeep it friendly and engaging.\nUnderstand their needs to see if our services align.\nIf there's a match, move to Step 2.\nStep 2: Ask Permission\nPolitely transition to a brief questionnaire via text.\nIf they're not ready, explore how else you can assist.\nStep 3: Service Questions\nIf they agree, proceed with the following questions, ensuring each is answered before moving to the next:\n\"Do you have a website for your business?\"\n\"Have you worked on improving your online reviews?\"\n\"Are you looking to expand your online reach beyond local areas?\"\nStep 4: Summarize Responses\nConfirm understanding: \"Based on your input, here's what I gathered. Does this sound right?\"\nAdjust as needed based on their feedback.\nStep 5: Book The Appointment\nIf they agree to a 30-minute consultation, ask if they prefer booking the appointment via text message, email, or phone.\nIf by text message: \"You’ll receive a text message with the link shortly.\"\nIf by email: \"Can I have your name and email address to send you the booking link?\"\nIf by phone: \"Okay, let me check the times available for you right now.\"\nFAQs:\nWhat services does Blinks Media offer? We specialize in web design, reputation management, and online marketing for local businesses to boost online visibility and growth.\nHow does your reputation management service work? We monitor, enhance, and promote positive customer reviews while addressing negative feedback constructively.\nCan you create a custom website for any type of business? Yes, we tailor web designs to meet each business's unique needs, ensuring functionality and aesthetic appeal.\nHow does Blinks Media help overcome geographical limitations? We use advanced online marketing strategies to help businesses reach a wider audience beyond their physical location.\nWhat's included in your free website mockup offer? Our free mockup showcases a proposed layout, user interface, and design theme for your business.\nHow long does it take to see results with your online marketing strategies? Initial results can appear in a few weeks, but significant growth typically takes a few months, depending on the strategy.\nDo you offer any guarantees on your services? While specific results can’t be guaranteed due to the dynamic digital landscape, we commit to delivering top-notch services to meet your goals.\nHow does Blinks Media differ from other digital service providers? We focus on custom solutions with expertise in both design and digital marketing, positioning us as a holistic partner for sustainable growth.\nHow can I monitor the progress of my campaign? We provide regular updates, analytics, and reports to keep you informed about your campaign's performance.\nAre there any long-term contracts? Our services are flexible. While some campaigns benefit from long-term approaches, we don't bind clients with restrictive contracts.\nCan you handle negative reviews and feedback for my business? Yes, our reputation management service proactively addresses negative feedback to improve public perception.\nDo you provide training or support for the websites you design? Yes, we offer training to help your team manage your new website effectively.\nHow do you determine the online marketing strategy for a business? We start with a thorough assessment of your business goals, target audience, and current online presence to guide our strategy.\nWhat platforms or tools do you use for online marketing? We use platforms like Google Ads, social media channels, SEO tools, and more, tailored to each business’s needs.\nHow can I get started with Blinks Media? Reach out through our contact form or call us. We’d love to discuss your business needs and how we can help.
${appConfig.enable_interruptions ? `
Interruption Handling:
- If the user interrupts, acknowledge it politely (e.g., "I see you have something to add," or "Oh, sorry for interrupting. You were saying?").
- After acknowledging, address their new input directly.
- Keep the acknowledgment brief and natural, then focus on responding to the new information or question.
` : ''}
`;

let chatHistory: ChatMessage[] = [
  { role: 'system', content: SYSTEM_MESSAGE }
];

// =====================================
// 4. Utility Functions
// =====================================
function delay(ms: number): Promise<void> {
  return new Promise(resolve => setTimeout(resolve, ms));
}

function log(message: string, data?: any) {
  const timestamp = new Date().toISOString();
  let logMessage = `${timestamp}: ${message}`;
  if (data) {
    logMessage += ` ${JSON.stringify(data)}`;
  }
  console.log(logMessage);
}

function switchLLMProvider(provider: LLMApiType) {
  CURRENT_LLM_API = provider;
  log('Switched LLM provider', { newProvider: provider });
}

function clearChatHistory() {
  chatHistory = [
    { role: 'system', content: SYSTEM_MESSAGE }
  ];
  log('Chat history cleared');
}

function sendClearMessage(ws: WebSocket, streamSid: string) {
  const clearMessage = {
    event: "clear",
    streamSid: streamSid
  };
  log('Sending clear message to Twilio', { message: clearMessage });
  ws.send(JSON.stringify(clearMessage));
}

// =====================================
// 5. LatencyTracker Class
// =====================================
class LatencyTracker {
  private latencies: { [key: string]: number[] } = {};
  private startTimes: { [key: string]: number } = {};

  startTracking(service: string) {
    this.startTimes[service] = Date.now();
  }

  endTracking(service: string) {
    const endTime = Date.now();
    const startTime = this.startTimes[service];
    if (startTime) {
      const latency = endTime - startTime;
      if (!this.latencies[service]) {
        this.latencies[service] = [];
      }
      this.latencies[service].push(latency);
      delete this.startTimes[service];
      this.logLatency(service, latency);
    }
  }

  private logLatency(service: string, latency: number) {
    console.log(`${service} Latency: ${latency}ms`);
  }

  getAverages() {
    const averages: { [key: string]: number } = {};
    for (const [service, latencies] of Object.entries(this.latencies)) {
      averages[service] = latencies.reduce((a, b) => a + b, 0) / latencies.length;
    }
    return averages;
  }

  logSummary() {
    const averages = this.getAverages();
    console.log('Call Latency Summary', { averageLatencies: averages });
  }

  getLatency(service: string): number | undefined {
    const latencies = this.latencies[service];
    if (latencies && latencies.length > 0) {
      return latencies[latencies.length - 1];
    }
    return undefined;
  }
}

// =====================================
// 6. TranscriptionService Class
// =====================================
class TranscriptionService {
  private dgConnection: ReturnType<DeepgramClient['listen']['live']>;
  private keepAliveInterval: NodeJS.Timeout | null = null;
  private reconnectTimeout: NodeJS.Timeout | null = null;
  private finalResult: string = '';
  private speechFinal: boolean = false;
  private isActive: boolean = true;
  private reconnectAttempts: number = 0;
  private maxReconnectAttempts: number = 5;
  private transcriptionStartTime: number = 0;
  private latencyTracker: LatencyTracker;
  private interruptionDetected: boolean = false;
  private enableInterruptions: boolean = false;
  private ws: WebSocket;
  private streamSid: string;

  constructor(
    private handleTranscription: (transcription: string, wasInterrupted: boolean) => void,
    latencyTracker: LatencyTracker,
    enable_interruptions: boolean = false,
    ws: WebSocket,
    streamSid: string
  ) {
    this.latencyTracker = latencyTracker;
    this.enableInterruptions = enable_interruptions;
    this.dgConnection = this.initializeConnection();
    this.ws = ws;
    this.streamSid = streamSid;
  }

  private initializeConnection(): ReturnType<DeepgramClient['listen']['live']> {
    const connection = deepgram.listen.live({
      encoding: 'mulaw',
      sample_rate: 8000,
      model: appConfig.transcriber_model,
      smart_format: true,
      punctuate: true,
      interim_results: true,
      endpointing: 200,
      utterance_end_ms: 1000,
      vad_events: true
    });

    this.setupEventListeners(connection);
    this.startKeepAlive();

    return connection;
  }

  private setupEventListeners(connection: ReturnType<DeepgramClient['listen']['live']>) {
    connection.on(LiveTranscriptionEvents.Open, () => {
      log('Deepgram connection opened');
      this.reconnectAttempts = 0;
    });

    connection.on(LiveTranscriptionEvents.Close, () => {
      log('Deepgram connection closed');
      this.stopKeepAlive();
      this.attemptReconnection();
    });

    connection.on(LiveTranscriptionEvents.Transcript, this.handleTranscriptEvent.bind(this));

    connection.on(LiveTranscriptionEvents.Error, (error) => {
      log('Deepgram error:', error);
    });
  }

  private async handleTranscriptEvent(transcriptionEvent: any) {
    const alternatives = transcriptionEvent.channel?.alternatives;
    let text = alternatives?.[0]?.transcript || '';

    if (transcriptionEvent.type === 'UtteranceEnd') {
      if (!this.speechFinal) {
        log(`UtteranceEnd received before speechFinal, emit the text collected so far: ${this.finalResult}`);
        this.latencyTracker.endTracking('Deepgram');
        this.handleTranscriptionWithInterruption(this.finalResult.trim());
        this.finalResult = '';
      } else {
        log('Speech was already final when UtteranceEnd received');
      }
      this.speechFinal = false;
      return;
    }

    if (transcriptionEvent.is_final && text.trim().length > 0) {
      this.finalResult += ` ${text}`;
      if (transcriptionEvent.speech_final || (this.enableInterruptions && this.interruptionDetected)) {
        this.speechFinal = true;
        this.latencyTracker.endTracking('Deepgram');
        this.handleTranscriptionWithInterruption(this.finalResult.trim());
        this.finalResult = '';
      } else {
        this.speechFinal = false;
        log('Interim final transcription:', { text: this.finalResult.trim() });
      }
    } else if (!transcriptionEvent.is_final && text.trim().length > 0) {
      log('Interim transcription:', { text: text.trim() });
      if (this.enableInterruptions) {
        await this.checkForInterruption(text);
      }
    }

    this.transcriptionStartTime = Date.now();
  }

  private handleTranscriptionWithInterruption(transcription: string) {
    const wasInterrupted = this.enableInterruptions && this.interruptionDetected;
    this.handleTranscription(transcription, wasInterrupted);
    this.interruptionDetected = false;
  }

  private async checkForInterruption(text: string) {
    log('Checking for interruption', {
      enableInterruptions: this.enableInterruptions,
      textLength: text.trim().length,
      isSpeaking: isSpeaking,
      wordToInterruptEnable: appConfig.wordToInterruptEnable,
      numberofwordstointeruptenable: appConfig.numberofwordstointeruptenable,
      numberofwordstointerupt: appConfig.words_to_interrupt
    });
    if (this.enableInterruptions && text.trim().length > 0 && isSpeaking) {
      let shouldInterrupt = false;

      if (appConfig.wordToInterruptEnable) {
        const lowerText = text.toLowerCase();
        shouldInterrupt = appConfig.wordsToInterrupt.some(word => lowerText.includes(word.toLowerCase()));
      }

     if (appConfig.numberofwordstointeruptenable) {
       const wordCount = text.trim().split(/\s+/).length;
       shouldInterrupt = shouldInterrupt || wordCount >= appConfig.words_to_interrupt;
       }
      
      if (shouldInterrupt) {
        log('Interruption detected', { text });
        this.interruptionDetected = true;
        
        // Send clear message immediately
        sendClearMessage(this.ws, this.streamSid);
        
        // Add delay here
        log(`Waiting for ${appConfig.clear_message_delay}ms after sending clear message`);
        await delay(appConfig.clear_message_delay);
        
        isSpeaking = false;
        interruptionEmitter.emit('interrupt', text);
      }
    }
  }

  private startKeepAlive() {
    this.stopKeepAlive();
    this.keepAliveInterval = setInterval(() => {
      if (this.dgConnection.getReadyState() === 1) {
        this.dgConnection.send(JSON.stringify({ type: 'KeepAlive' }));
        log('Sent KeepAlive message');
      }
    }, 3000);
  }

  private stopKeepAlive() {
    if (this.keepAliveInterval) {
      clearInterval(this.keepAliveInterval);
      this.keepAliveInterval = null;
    }
  }

  private attemptReconnection() {
    if (this.isActive && this.reconnectAttempts < this.maxReconnectAttempts) {
      this.reconnectAttempts++;
      log(`Attempting to reconnect (${this.reconnectAttempts}/${this.maxReconnectAttempts})`);
      this.reconnectTimeout = setTimeout(() => {
        this.dgConnection = this.initializeConnection();
      }, 5000 * this.reconnectAttempts);
    } else {
      log('Max reconnection attempts reached or service is no longer active');
    }
  }

  public send(payload: string) {
    if (this.dgConnection.getReadyState() === 1) {
      this.latencyTracker.startTracking('Deepgram');
      this.dgConnection.send(Buffer.from(payload, 'base64'));
    } else {
      log('Cannot send: Deepgram connection is not open');
      this.latencyTracker.endTracking('Deepgram'); // End tracking if send fails
    }
  }

  public close() {
    this.isActive = false;
    this.stopKeepAlive();
    if (this.dgConnection.getReadyState() === 1) {
      this.dgConnection.finish();
      log('Deepgram connection finished');
    }
    this.latencyTracker.endTracking('Deepgram'); // Ensure tracking ends if connection closes
    log('TranscriptionService closed');
  }

  public setEnableInterruptions(enable: boolean) {
    this.enableInterruptions = enable;
  }
}

let lastResetTime = Date.now();

const resetSilenceTimer = () => {
  if (silenceTimer) {
    clearTimeout(silenceTimer);
  }
  if (!isWaitingForAudioComplete) {
    silenceTimer = setTimeout(() => {
      log('Silence timeout reached');
      if (ws && ws.readyState === WebSocket.OPEN) {
        ws.close();
      }
    }, appConfig.silenceTimeoutSeconds * 1000);
    log(`Silence timer started with timeout: ${appConfig.silenceTimeoutSeconds} seconds`);
  } else {
    log('Silence timer not started, waiting for audio_complete');
  }
};

// =====================================
// 7. Main Application Logic
// =====================================
async function startApp() {
  try {
    await fetchAssistantSettings(); // Remove the argument if it's present
  } catch (error) {
    log('Failed to fetch assistant settings:', { error: (error as Error).message });
    // You might want to implement a retry mechanism or exit the application here
    process.exit(1); // Exit the application if we can't fetch the settings
  }

  app.use(express.json());
  app.use(express.urlencoded({ extended: true }));

  app.use('/webhook', webhookRouter);

  app.post('/call/incoming', (req, res) => {
    calledNumber = req.body.Called;
    log('Incoming call received', { body: req.body, calledNumber });

    const twimlResponse = new twiml.VoiceResponse();
    twimlResponse.connect().stream({
      url: `wss://${process.env.SERVER_DOMAIN}:8443/call/connection`,
    });
  
    log('TwiML response created');

    const responseXml = twimlResponse.toString();
    res.type('text/xml').send(responseXml);
    log('TwiML response sent', { response: responseXml, calledNumber });
  });

  app.ws('/call/connection', async (websocket: WebSocket) => {
    ws = websocket; // Assign the websocket to the global ws variable
    log('WebSocket connection established');
    let streamSid: string | undefined;
    let conversation: ConversationMessage[] = [];
    let latencyTracker = new LatencyTracker();
    let transcriptionService: TranscriptionService | null = null;
    isFirstMessage = true; // Reset isFirstMessage for each new connection
  
    const handleTranscription = async (transcription: string, wasInterrupted: boolean) => {
      log('Transcription received', { transcription, wasInterrupted });
      if (!isProcessing && streamSid) {
        isProcessing = true;
        if (isFirstMessage) {
          isFirstMessage = false;
          log('First message received');
        }
        if (silenceTimer) {
          clearTimeout(silenceTimer);
          silenceTimer = null;
          log('Silence timer stopped due to transcription');
        }
        isWaitingForAudioComplete = true;
        await handleUserInput(ws, streamSid, transcription, conversation, latencyTracker, wasInterrupted);
        isProcessing = false;
      }
    };

    // Don't start the timer when the connection is established
  
    interruptionEmitter.on('interrupt', async (interruptionText) => {
      isSpeaking = false; // Stop the current speech
      isAssistantSpeaking = false; // Update assistant speaking state
      log('is Speaking', { isSpeaking: isSpeaking, isAssistantSpeaking: isAssistantSpeaking });
      log('Interruption detected', { interruptionText });
      if (streamSid) {
        isProcessing = true;
        await handleUserInput(ws, streamSid, interruptionText, conversation, latencyTracker, true);
        isProcessing = false;
      }
    });

    ws.on('message', async (data) => {
      let message: TwilioMessage;

      try {
        message = JSON.parse(data.toString()) as TwilioMessage;
      } catch (error) {
        log('Error parsing WebSocket message', { error: (error as Error).message });
        return;
      }

      switch (message.event) {
        case 'connected':
          log('Connected event received');
          streamSid = message.streamSid;
          break;

        case 'start':
          if (message.start) {
            log('Start event received', { streamSid: message.start.streamSid });
            streamSid = message.start.streamSid;
          
            log('Current phone number:', calledNumber || 'Not available');
          
            if (calledNumber) {
              const assistantSettings = await getAssistantSettings(calledNumber);
              if (assistantSettings) {
                if (assistantSettings.silence_timeout) {
                  appConfig.silenceTimeoutSeconds = parseInt(assistantSettings.silence_timeout);
                  log(`Applied silence timeout: ${appConfig.silenceTimeoutSeconds} seconds`);
                }
                if (assistantSettings.model_provider) {
                  setLLMApiType(assistantSettings.model_provider);
                  log(`Applied model provider: ${assistantSettings.model_provider}`);
                }
                if (assistantSettings.maximum_call_duration_enable !== undefined) {
                  appConfig.maximum_call_duration_enable = assistantSettings.maximum_call_duration_enable;
                }
                if (assistantSettings.maximum_call_duration) {
                  appConfig.maximum_call_duration = parseInt(assistantSettings.maximum_call_duration);
                }
                log(`Applied maximum call duration settings: enabled=${appConfig.maximum_call_duration_enable}, duration=${appConfig.maximum_call_duration} seconds`);
              } else {
                log(`No specific settings found for ${calledNumber}. Using default settings.`);
              }
            } else {
              log('No phone number available. Using default settings.');
            }

            log('Full start message:', JSON.stringify(message.start, null, 2));
          
            transcriptionService = new TranscriptionService(
              handleTranscription, 
              latencyTracker, 
              appConfig.enable_interruptions,
              ws,
              streamSid
            );
            await processInitialGreeting(ws, streamSid, latencyTracker);

            // Start the maximum call duration timer
            handleMaxCallDuration(ws, streamSid);
          }
          break;

        case 'media':
          if (message.media && streamSid && transcriptionService) {
            transcriptionService.send(message.media.payload);
            // Don't reset the timer here
          }
          break;

        case 'stop':
          log('Stop event received - Call ending');
          if (transcriptionService) {
            transcriptionService.close();
            transcriptionService = null;
            log('Deepgram connection closed due to call ending');
          }
          clearChatHistory();
          latencyTracker.logSummary();
          calledNumber = ''; // Reset the calledNumber
          ws.close();
          break;

        case 'dtmf':
          if (message.dtmf) {
            log('DTMF event received', { digit: message.dtmf.digit });
          }
          break;

        case 'mark':
          if (message.mark && message.mark.name === 'audio_complete') {
            log('Received audio_complete mark from Twilio');
            isSpeaking = false;
            isAssistantSpeaking = false;
            isWaitingForAudioComplete = false;
            log('Speech playback completed', { isSpeaking: isSpeaking, isAssistantSpeaking: isAssistantSpeaking, timestamp: Date.now() });
            log('Starting silence timer');
            resetSilenceTimer();
          }
          break;

        default:
          log('Unhandled WebSocket event', { event: message.event });
      }
    });

    ws.on('error', (error) => {
      log('WebSocket error', { error: (error as Error).message });
      if (transcriptionService) {
        transcriptionService.close();
        transcriptionService = null;
        log('Deepgram connection closed due to WebSocket error');
      }
      clearChatHistory();
    });

    ws.on('close', (code, reason) => {
      log('WebSocket closed', { code, reason: reason || 'No reason provided' });
      if (transcriptionService) {
        transcriptionService.close();
        transcriptionService = null;
        log('Deepgram connection closed due to WebSocket closure');
      }
      if (silenceTimer) {
        clearTimeout(silenceTimer);
        silenceTimer = null;
      }
      clearChatHistory();
      interruptionEmitter.removeAllListeners('interrupt'); // Clean up the listener when the connection closes
    });
  });

  app.use((err: any, req: express.Request, res: express.Response, next: express.NextFunction) => {
    log('Error occurred', { error: err.stack });
    res.status(500).send('Something broke!');
  });

  app.listen(PORT, '0.0.0.0', () => {
    log('Server started', { port: PORT });
  });
}

// =====================================
// 8. Helper Functions
// =====================================
function handleMaxCallDuration(ws: WebSocket, streamSid: string) {
  if (appConfig.maximum_call_duration_enable && appConfig.maximum_call_duration > 0) {
    setTimeout(() => {
      log('Maximum call duration reached. Ending call.');
      textToSpeechAndStream(ws, streamSid, appConfig.end_call_message, new LatencyTracker())
        .then(() => {
          if (ws.readyState === WebSocket.OPEN) {
            ws.close();
          }
        })
        .catch((error) => {
          log('Error ending call due to maximum duration', { error: (error as Error).message });
          if (ws.readyState === WebSocket.OPEN) {
            ws.close();
          }
        });
    }, appConfig.maximum_call_duration * 1000);
  }
}

function getValidOpenAIVoice(voice: string): OpenAIVoice {
  const validVoices: OpenAIVoice[] = ["alloy", "echo", "fable", "onyx", "nova", "shimmer"];
  if (validVoices.includes(voice as OpenAIVoice)) {
    return voice as OpenAIVoice;
  }
  console.warn(`Invalid OpenAI voice: ${voice}. Defaulting to 'alloy'.`);
  return "alloy";
}

async function processInitialGreeting(ws: WebSocket, streamSid: string, latencyTracker: LatencyTracker) {
  try {
    isProcessing = true;
    log('Processing initial greeting');
    latencyTracker.startTracking('LLM');
    const openAIResponse = await getLLMResponse("Hi", null, latencyTracker, CURRENT_LLM_API);
    latencyTracker.endTracking('LLM');
    log('LLM response received', { response: openAIResponse });
    await textToSpeechAndStream(ws, streamSid, openAIResponse, latencyTracker);
  } catch (error) {
    log('Error in processInitialGreeting', { error: (error as Error).message });
    await textToSpeechAndStream(ws, streamSid, "I'm sorry, but I encountered an error. Please try again later.", latencyTracker);
  } finally {
    isProcessing = false;
  }
}

async function handleUserInput(ws: WebSocket, streamSid: string, userInput: string, conversation: ConversationMessage[], latencyTracker: LatencyTracker, wasInterrupted: boolean) {
  latencyTracker.startTracking('Total');
  try {
    log('Handling user input', { userInput });
    
    // Stop the silence timer when user input is received
    isWaitingForAudioComplete = true;
    resetSilenceTimer();
    
    let contextualPrompt = userInput;
    if (appConfig.enable_interruptions && isAssistantSpeaking) {
      contextualPrompt = `The user interrupted with: "${userInput}". Please acknowledge the interruption politely and address their new input.`;
    }
    
    conversation.push({ role: 'user', content: contextualPrompt });
    log('Current conversation context', { conversation });

    // Implement LLM request delay
    log(`Waiting for ${appConfig.llm_request_delay} seconds before sending request to LLM`);
    await delay(appConfig.llm_request_delay * 1000);

    latencyTracker.startTracking('LLM');
    const openAIResponse = await getLLMResponse(contextualPrompt, currentThreadId, latencyTracker, CURRENT_LLM_API);
    latencyTracker.endTracking('LLM');

    log('LLM response received', { response: openAIResponse });
    if (typeof openAIResponse === 'string') {
      conversation.push({ role: 'assistant', content: openAIResponse });
    }

    // Implement response delay
    log(`Waiting for ${appConfig.response_delay} seconds before responding`);
    await delay(appConfig.response_delay * 1000);

    isAssistantSpeaking = true;
    latencyTracker.startTracking('TTS');
    await textToSpeechAndStream(ws, streamSid, openAIResponse, latencyTracker);
    latencyTracker.endTracking('TTS');

    latencyTracker.endTracking('Total');

    log('Latency Breakdown', {
      deepgram: latencyTracker.getLatency('Deepgram'),
      llm: latencyTracker.getLatency('LLM'),
      tts: latencyTracker.getLatency('TTS'),
      total: latencyTracker.getLatency('Total')
    });

    // Don't reset the silence timer here, it will be reset when 'audio_complete' is received
  } catch (error) {
    log('Error in handleUserInput', { error: (error as Error).message });
    await textToSpeechAndStream(ws, streamSid, "I'm sorry, but I encountered an error. Could you please try again?", latencyTracker);
    isAssistantSpeaking = false;
    resetSilenceTimer();
  }
}

async function getLLMResponse(
  input: string, 
  threadId: string | null, 
  latencyTracker: LatencyTracker, 
  apiType: LLMApiType = CURRENT_LLM_API
): Promise<string | Readable> {
  if (Array.isArray(input)) {
    input = input.join(' ');
  }
  log('Sending message to LLM', { input, threadId, apiType });

  latencyTracker.startTracking('LLM');
  try {
    const assistantSettings = await getAssistantSettings(calledNumber);
    if (apiType === LLMApiType.OpenAIAssistant) {
      let thread;
      if (threadId) {
        thread = await openai.beta.threads.retrieve(threadId);
        await openai.beta.threads.messages.create(threadId, {
          role: 'user',
          content: input
        });
      } else {
        thread = await openai.beta.threads.create({
          messages: [{ role: 'user', content: input }]
        });
        currentThreadId = thread.id;
      }

      log('Using thread', { threadId: thread.id });

      const stream = new Readable({
        read() {}
      });

      const runStream = await openai.beta.threads.runs.createAndStream(thread.id, {
        assistant_id: process.env.OPENAI_ASSISTANT_ID!,
      });

      (async () => {
        for await (const chunk of runStream) {
          const event = chunk as AssistantStreamEvent;
          if (event.event === 'thread.message.delta') {
            const delta = event.data.delta;
            if (delta.content && delta.content[0].type === 'text') {
              const chunkValue = delta.content[0].text.value;
              log('Received chunk from OpenAI Assistant', { chunk: chunkValue });
              stream.push(chunkValue);
            }
          }
        }
        stream.push(null);
      })();

      return stream;
    } else if (apiType === LLMApiType.OpenAIChat) {
      chatHistory.push({ role: 'user', content: input });

      switch (COMMUNICATION_METHOD) {
        case CommunicationMethod.WebSocket: {
          const stream = new Readable({
            read() {}
          });

          const chatStream = await openai.chat.completions.create({
            model: assistantSettings.model || "gpt-4o-mini-2024-07-18",
            messages: chatHistory,
            stream: true,
            temperature: assistantSettings.temperature || 1,
            max_tokens: assistantSettings.max_tokens || 256,
            top_p: assistantSettings.top_p || 1,
            frequency_penalty: 0,
            presence_penalty: 0,
          });

          let assistantResponse = '';

          (async () => {
            for await (const chunk of chatStream) {
              if (chunk.choices[0]?.delta?.content) {
                const chunkValue = chunk.choices[0].delta.content;
                log('Received chunk from OpenAI Chat', { chunk: chunkValue });
                stream.push(chunkValue);
                assistantResponse += chunkValue;
              }
            }
            chatHistory.push({ role: 'assistant', content: assistantResponse });
            
            if (chatHistory.length > 10) {
              chatHistory = [
                chatHistory[0],
                ...chatHistory.slice(-9)
              ];
            }
            
            stream.push(null);
          })();

          return stream;
        }
        case CommunicationMethod.Streaming:
        case CommunicationMethod.NonStreaming:
        default: {
          latencyTracker.startTracking('LLM');    
          const completion = await openai.chat.completions.create({
            model: assistantSettings.model || "gpt-4o-mini-2024-07-18",
            messages: chatHistory,
            temperature: assistantSettings.temperature || 1,
            max_tokens: assistantSettings.max_tokens || 256,
            top_p: assistantSettings.top_p || 1,
            frequency_penalty: 0,
            presence_penalty: 0,
          });
          latencyTracker.endTracking('LLM');
          
          const assistantResponse = completion.choices[0]?.message?.content || '';
          chatHistory.push({ role: 'assistant', content: assistantResponse });
          
          if (chatHistory.length > 10) {
            chatHistory = [
              chatHistory[0],
              ...chatHistory.slice(-9)
            ];
          }

          return assistantResponse;
        }
      }
    } else if (apiType === LLMApiType.Groq) {
      chatHistory.push({ role: 'user', content: input });

      switch (COMMUNICATION_METHOD) {
        case CommunicationMethod.WebSocket: {
          const stream = new Readable({
            read() {}
          });

          const chatStream = await groq.chat.completions.create({
            model: assistantSettings.model || "llama3-groq-70b-8192-tool-use-preview",
            messages: chatHistory,
            stream: true,
            temperature: assistantSettings.temperature || 0.65,
            max_tokens: assistantSettings.max_tokens || 256,
            top_p: assistantSettings.top_p || 1,
          });

          let assistantResponse = '';

          (async () => {
            for await (const chunk of chatStream) {
              if (chunk.choices[0]?.delta?.content) {
                const chunkValue = chunk.choices[0].delta.content;
                log('Received chunk from Groq', { chunk: chunkValue });
                stream.push(chunkValue);
                assistantResponse += chunkValue;
              }
            }
            chatHistory.push({ role: 'assistant', content: assistantResponse });
            
            if (chatHistory.length > 10) {
              chatHistory = [
                chatHistory[0],
                ...chatHistory.slice(-9)
              ];
            }
            
            stream.push(null);
          })();

          return stream;
        }
        case CommunicationMethod.Streaming:
        case CommunicationMethod.NonStreaming:
        default: {
          latencyTracker.startTracking('LLM');    
          const completion = await groq.chat.completions.create({
            model: assistantSettings.model || "llama3-groq-70b-8192-tool-use-preview",
            messages: chatHistory,
            stream: false,
            temperature: assistantSettings.temperature || 0.65,
            max_tokens: assistantSettings.max_tokens || 256,
            top_p: assistantSettings.top_p || 1,
          });
          latencyTracker.endTracking('LLM');
          
          const assistantResponse = completion.choices[0]?.message?.content || '';
          chatHistory.push({ role: 'assistant', content: assistantResponse });
          
          if (chatHistory.length > 10) {
            chatHistory = [
              chatHistory[0],
              ...chatHistory.slice(-9)
            ];
          }

          return assistantResponse;
        }
      }
    } else {
      throw new Error(`Unsupported LLM API type: ${apiType}`);
    }
  } catch (error) {
    latencyTracker.endTracking('LLM');
    throw error;
  }
}

// =====================================
// 9. Text-to-Speech and Streaming Functions
// =====================================
async function textToSpeechAndStream(ws: WebSocket, streamSid: string, text: string | Readable, latencyTracker: LatencyTracker) {
  let interruptionOccurred = false;
  let audioCompleteSent = false;
  const handleInterruption = (ws: WebSocket, streamSid: string) => {
    if (isSpeaking) {
      interruptionOccurred = true;
      log('Interruption detected, sending clear message');
      sendClearMessage(ws, streamSid);
      // Note: We'll set isSpeaking to false after receiving confirmation from Twilio
    }
  };
  try {
    log('Initializing TTS API');
    
    let textToConvert: string;
    if (text instanceof Readable) {
      textToConvert = await new Promise<string>((resolve, reject) => {
        let content = '';
        text.on('data', (chunk) => content += chunk);
        text.on('end', () => resolve(content));
        text.on('error', reject);
      });
    } else {
      textToConvert = text;
    }
    log('Converting text to speech', { textLength: textToConvert.length });

    // Set up interruption listener just before starting to speak
    interruptionEmitter.on('interrupt', () => handleInterruption(ws, streamSid));
    isSpeaking = true;
    log('Is Speaking', { isSpeaking: isSpeaking });
    log('Starting speech', { timestamp: Date.now() });
    
    latencyTracker.startTracking('TTS');
    
    switch (appConfig.voice_provider.toLowerCase()) {
	 case 'elevenlabs':
			switch (COMMUNICATION_METHOD) {
			  case CommunicationMethod.WebSocket: {
				const elevenlabsWs = new WebSocket(`wss://api.elevenlabs.io/v1/text-to-speech/${appConfig.voice_id}/stream-input?model_id=eleven_turbo_v2&output_format=ulaw_8000&optimize_streaming_latency=3`);

				elevenlabsWs.on('open', async () => {
				  log('ElevenLabs WebSocket opened');
				  elevenlabsWs.send(JSON.stringify({
					"text": " ",
					"voice_settings": {
					  "stability": appConfig.stability,
					  "similarity_boost": appConfig.clarity_boost
					},
					"xi_api_key": process.env.ELEVENLABS_API_KEY
				  }));

				  const sendChunk = (chunk: string) => {
					log('Sending chunk to ElevenLabs', { chunk });    
					elevenlabsWs.send(JSON.stringify({ 
					  "text": chunk,
					}));
				  };

				  const processText = async () => {
					if (text instanceof Readable) {
					  for await (const chunk of text_chunker(createAsyncIterableFromReadable(text))) {
						if (interruptionOccurred) break;
						sendChunk(chunk);
					  }
					} else {
					  for await (const chunk of text_chunker(createAsyncIterableFromString(text))) {
						if (interruptionOccurred) break;
						sendChunk(chunk);
					  }
					}
					elevenlabsWs.send(JSON.stringify({ "flush": true }));
					elevenlabsWs.send(JSON.stringify({ "text": "" }));
					log('Sending Final Messages to ElevenLabs'); 
				  };

				  processText();
				});

				let isFirstChunk = true;

				elevenlabsWs.on('message', (data) => {
				  const message = JSON.parse(data.toString());
				  if (message.audio) {
					const audioChunk = Buffer.from(message.audio, 'base64');
					log('Received audio chunk from ElevenLabs', { 
					  chunkSizeBytes: audioChunk.length,
					  firstFewBytes: audioChunk.slice(0, 10).toString('hex')
					});
					if (!interruptionOccurred) {
					  sendMediaMessage(ws, streamSid, audioChunk);
					}
					
					if (isFirstChunk) {
					  isFirstChunk = false;
					  log('Received first audio chunk from ElevenLabs');
					}
				  } else if (message.isFinal) {
					log('Received final message from ElevenLabs');
					if (!interruptionOccurred && !audioCompleteSent) {
					  audioCompleteSent = true;
					  sendMarkMessage(ws, streamSid, 'audio_complete');
					}
				  }
				});

				elevenlabsWs.on('error', (error) => {
				  log('ElevenLabs WebSocket error', { error: error.message });
				  if (!audioCompleteSent) {
					audioCompleteSent = true;
					sendMarkMessage(ws, streamSid, 'audio_complete');
				  }
				});

				elevenlabsWs.on('close', () => {
				  log('ElevenLabs WebSocket closed');
				});

				await new Promise((resolve) => {
				  elevenlabsWs.on('close', resolve);
				});

				break;
			  }

			  case CommunicationMethod.Streaming:
				const audioStream = await elevenlabs.generate({
				  text: textToConvert,
				  voice: appConfig.voice_id,
				  model_id: modelId,
				  voice_settings: {
					stability: appConfig.stability,
					similarity_boost: appConfig.clarity_boost
				  },
				  output_format: 'ulaw_8000',
				  optimize_streaming_latency: '3'
				});
			  
				log('Starting audio streaming to Twilio');
				
				let isFirstChunk = true;
				audioStream.on('data', (chunk) => {
				  if (isFirstChunk) {
					latencyTracker.endTracking('TTS');
					isFirstChunk = false;
				  }
				  log('Received audio chunk from ElevenLabs', { size: chunk.length });
				  if (!interruptionOccurred) {
					sendMediaMessage(ws, streamSid, chunk);
				  }
				});
			  
				audioStream.on('end', () => {
				  log('ElevenLabs audio stream ended');
				  if (!interruptionOccurred && !audioCompleteSent) {
					audioCompleteSent = true;
					sendMarkMessage(ws, streamSid, 'audio_complete');
				  }
				});
			  
				await new Promise((resolve) => audioStream.on('end', resolve));
				break;

			  case CommunicationMethod.NonStreaming:
			  default:
				const audioResponse = await elevenlabs.textToSpeech.convert(appConfig.voice_id, {
				  text: textToConvert,
				  model_id: modelId,
				  voice_settings: {
					stability: 0.5,
					similarity_boost: 0.75
				  },
				  output_format: 'ulaw_8000',
				  optimize_streaming_latency: '3'
				});
				
				latencyTracker.endTracking('TTS');

				log('Received response from ElevenLabs', { 
				  responseType: typeof audioResponse, 
				  isBuffer: Buffer.isBuffer(audioResponse),
				  isArrayBuffer: audioResponse instanceof ArrayBuffer,
				  isReadable: audioResponse instanceof Readable
				});

				let audioBuffer: Buffer;
				if (audioResponse instanceof ArrayBuffer) {
				  audioBuffer = Buffer.from(audioResponse);
				} else if (Buffer.isBuffer(audioResponse)) {
				  audioBuffer = audioResponse;
				} else if (audioResponse instanceof Readable) {
				  audioBuffer = await new Promise<Buffer>((resolve, reject) => {
					const chunks: Buffer[] = [];
					audioResponse.on('data', (chunk) => chunks.push(Buffer.from(chunk)));
					audioResponse.on('end', () => resolve(Buffer.concat(chunks)));
					audioResponse.on('error', reject);
				  });
				} else {
				  throw new Error(`Unexpected response type from ElevenLabs: ${typeof audioResponse}`);
				}

				log('Processed audio from ElevenLabs', { size: audioBuffer.length });
				if (!interruptionOccurred) {
				  sendMediaMessage(ws, streamSid, audioBuffer);
				  audioCompleteSent = true;
				  sendMarkMessage(ws, streamSid, 'audio_complete');
				}
				break;
			}
			break;

		  case 'openai':
			log('Starting OpenAI TTS process');
			const mp3Audio = await openai.audio.speech.create({
			  model: "tts-1",
			  voice: getValidOpenAIVoice(appConfig.voice_id),
			  input: textToConvert,
			  response_format: "mp3",
			});

			log('Converting OpenAI response to audio stream');
			
			// Convert the response to a readable stream
			const audioStream = new Readable({
			  read() {}
			});

			// Process the audio data
			(async () => {
			  const audioResponse = mp3Audio as unknown as { arrayBuffer: () => Promise<ArrayBuffer> };
			  const buffer = Buffer.from(await audioResponse.arrayBuffer());
			  audioStream.push(buffer);
			  audioStream.push(null); // Signal the end of the stream
			})();

			log('Starting FFmpeg conversion process');

			const ffmpegProcess = spawn(ffmpegPath, [
			  '-i', 'pipe:0',
			  '-acodec', 'pcm_mulaw',
			  '-ar', '8000',
			  '-f', 'mulaw',
			  'pipe:1'
			]);

			audioStream.pipe(ffmpegProcess.stdin);

			let isFirstChunk = true;
			ffmpegProcess.stdout.on('data', (chunk) => {
			  if (isFirstChunk) {
				latencyTracker.endTracking('TTS');
				isFirstChunk = false;
			  }
			  log('Sending converted audio chunk to Twilio', { size: chunk.length });
			  if (!interruptionOccurred) {
				sendMediaMessage(ws, streamSid, chunk);
			  }
			});

			ffmpegProcess.stderr.on('data', (data) => {
			  log('FFmpeg stderr:', data.toString());
			});

			ffmpegProcess.on('close', (code) => {
			  log('FFmpeg process closed', { code });
			  if (!interruptionOccurred && !audioCompleteSent) {
				audioCompleteSent = true;
				sendMarkMessage(ws, streamSid, 'audio_complete');
			  }
			});

			await new Promise((resolve) => ffmpegProcess.on('close', resolve));
			break;
		  
		case 'deepgram':
	  log('Starting Deepgram TTS process');
	  
	  try {
		const response = await deepgram.speak.request(
		  { text: textToConvert },
		  {
			model: appConfig.voice_id,
			encoding: "mulaw",
			sample_rate: 8000
		  }
		);
		const stream = await response.getStream();
		
		if (stream) {
		  log('Received audio stream from Deepgram');
		  latencyTracker.endTracking('TTS');
		  const reader = stream.getReader();
		  
		  while (true) {
			const { done, value } = await reader.read();
			
			if (done) {
			  log('Deepgram stream completed naturally');
			  break;
			}
			sendMediaMessage(ws, streamSid, Buffer.from(value));
		  }
		  
		  log('Sending audio_complete mark');
		  sendMarkMessage(ws, streamSid, 'audio_complete');
		} else {
		  throw new Error('No audio stream received from Deepgram');
		}
	  } catch (deepgramError) {
		log('Error in Deepgram TTS', { error: (deepgramError as Error).message });
		throw deepgramError;
	  }
	  break;
      case 'rime_ai':
        log('Starting RimeAI TTS process');
        try {
          const response = await fetch('https://users.rime.ai/v1/rime-tts', {
            method: 'POST',
            headers: {
              Accept: 'audio/x-mulaw',
              Authorization: `Bearer ${process.env.RIME_AI_API_KEY}`,
              'Content-Type': 'application/json'
            },
            body: JSON.stringify({
              speaker: appConfig.voice_id,
              text: textToConvert,
              modelId: RIME_AI_CONFIG.modelId,
              speedAlpha: appConfig.speed,
              reduceLatency: RIME_AI_CONFIG.reduceLatency
            })
          });

          if (!response.ok) {
            throw new Error(`RimeAI API error: ${response.status} ${response.statusText}`);
          }

          if (!response.body) {
            throw new Error('No response body received from RimeAI');
          }

          const reader = response.body.getReader();
          log('Received audio stream from RimeAI');
          latencyTracker.endTracking('TTS');

          while (true) {
            const { done, value } = await reader.read();
            
            if (done) {
              log('RimeAI stream completed naturally');
              break;
            }

            sendMediaMessage(ws, streamSid, Buffer.from(value));
            
            // Small delay to allow for interruption checking
            await new Promise(resolve => setTimeout(resolve, 50));

            if (interruptionOccurred) {
              log('Interruption occurred, stopping RimeAI audio stream');
              await reader.cancel();
              break;
            }
          }
          
          if (!interruptionOccurred) {
            log('Sending audio_complete mark');
            sendMarkMessage(ws, streamSid, 'audio_complete');
          }
        } catch (rimeAIError) {
          log('Error in RimeAI TTS', { error: (rimeAIError as Error).message });
          throw rimeAIError;
        }
        break;
    }
  } catch (error) {
    log('Error in textToSpeechAndStream', { error: (error as Error).message, stack: (error as Error).stack });
    const errorMessage = "I'm sorry, but I encountered an error while processing the audio. Please try again.";
    const errorBuffer = Buffer.from(errorMessage, 'utf-8');
    sendMediaMessage(ws, streamSid, errorBuffer);
    sendMarkMessage(ws, streamSid, 'audio_complete');
    latencyTracker.endTracking('TTS');
  } finally {
    log('Speech completed', { timestamp: Date.now() });
  }
}

async function* text_chunker(chunks: AsyncIterable<string>): AsyncGenerator<string> {
  const splitters = [".", ",", "?", "!", ";", ":", "—", "-", "(", ")", "[", "]", "}", " "];
  let buffer = "";

  for await (const text of chunks) {
    if (splitters.some(splitter => buffer.endsWith(splitter))) {
      yield buffer + " ";
      buffer = text;
    } else if (splitters.some(splitter => text.startsWith(splitter))) {
      yield buffer + text[0] + " ";
      buffer = text.slice(1);
    } else {
      buffer += text;
    }
  }

  if (buffer) {
    yield buffer + " ";
  }
}

function createAsyncIterableFromReadable(readable: Readable): AsyncIterable<string> {
  return {
    [Symbol.asyncIterator]() {
      return {
        next() {
          return new Promise((resolve) => {
            readable.once('data', (chunk) => {
              resolve({ value: chunk.toString(), done: false });
            });
            readable.once('end', () => {
              resolve({ value: undefined, done: true });
            });
          });
        },
      };
    },
  };
}

function createAsyncIterableFromString(str: string): AsyncIterable<string> {
  return {
    [Symbol.asyncIterator]() {
      let done = false;
      return {
        next() {
          if (done) {
            return Promise.resolve({ done: true, value: undefined });
          }
          done = true;
          return Promise.resolve({ done: false, value: str });
        },
      };
    },
  };
}

function sendMediaMessage(ws: WebSocket, streamSid: string, payload: Buffer) {
  const message = {
    event: 'media',
    streamSid: streamSid,
    media: {
      payload: payload.toString('base64')
    },
    sequenceNumber: (mediaSequenceNumber++).toString()
  };
  const messageString = JSON.stringify(message);
  log('Sending media message to Twilio', { 
    streamSid,
    sequenceNumber: message.sequenceNumber,
    payloadSize: payload.length, 
    messageSize: messageString.length
  });
  ws.send(messageString);
}

function sendMarkMessage(ws: WebSocket, streamSid: string, name: string) {
  const message = {
    event: 'mark',
    streamSid: streamSid,
    mark: {
      name: name
    }
  };
  log('Sending mark message', { message });
  ws.send(JSON.stringify(message));
}

// =====================================
// 10. Webhook Handler
// =====================================
app.post('/webhook', express.json(), async (req, res) => {
  const authHeader = req.headers.authorization;
  if (!authHeader || !authHeader.startsWith('Bearer ')) {
    log('Webhook request received without proper authorization');
    return res.status(401).json({ error: 'Unauthorized' });
  }

  const token = authHeader.split(' ')[1];
  if (token !== process.env.VOICE360_API_KEY) {
    log('Webhook request received with invalid API key');
    return res.status(401).json({ error: 'Invalid API key' });
  }

  try {
    await fetchAssistantSettings();
    log('Cache updated successfully via webhook');
    res.json({ message: 'Cache updated successfully' });
  } catch (error) {
    log('Error updating cache via webhook', { error: (error as Error).message });
    res.status(500).json({ error: 'Failed to update cache' });
  }
});

// =====================================
// 11. Error Handling and Process Event Listeners
// =====================================
process.on('unhandledRejection', (reason, promise) => {
  log('Unhandled Rejection', { reason });
});

process.on('uncaughtException', (error) => {
  log('Uncaught Exception', { error: (error as Error).message, stack: (error as Error).stack });
  process.exit(1);
});

// Application Start
log('Starting application');
startApp();
